"""
Type of SDN controllers in CSLE
"""

from enum import IntEnum


class SDNControllerType(IntEnum):
    """
    Enum representing the different SDN controller types supported in CSLE
    """
    RYU = 0
