from enum import IntEnum


class EmulationAttackerActionId(IntEnum):
    """
    Enum representing the different attacks in the emulation
    """
    TCP_SYN_STEALTH_SCAN_HOST = 0
    PING_SCAN_HOST = 1
    UDP_PORT_SCAN_HOST = 2
    TCP_CON_NON_STEALTH_SCAN_HOST = 3
    TCP_FIN_SCAN_HOST = 4
    TCP_NULL_SCAN_HOST = 5
    TCP_XMAS_TREE_SCAN_HOST = 6
    OS_DETECTION_SCAN_HOST = 7
    VULSCAN_HOST = 8
    NMAP_VULNERS_HOST = 9
    TELNET_SAME_USER_PASS_DICTIONARY_HOST = 10
    SSH_SAME_USER_PASS_DICTIONARY_HOST = 11
    FTP_SAME_USER_PASS_DICTIONARY_HOST = 12
    CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST = 13
    IRC_SAME_USER_PASS_DICTIONARY_HOST = 14
    MONGO_SAME_USER_PASS_DICTIONARY_HOST = 15
    MYSQL_SAME_USER_PASS_DICTIONARY_HOST = 16
    SMTP_SAME_USER_PASS_DICTIONARY_HOST = 17
    POSTGRES_SAME_USER_PASS_DICTIONARY_HOST = 18
    TCP_SYN_STEALTH_SCAN_ALL = 19
    PING_SCAN_ALL = 20
    UDP_PORT_SCAN_ALL = 21
    TCP_CON_NON_STEALTH_SCAN_ALL = 22
    TCP_FIN_SCAN_ALL = 23
    TCP_NULL_SCAN_ALL = 24
    TCP_XMAS_TREE_SCAN_ALL = 25
    OS_DETECTION_SCAN_ALL = 26
    VULSCAN_ALL = 27
    NMAP_VULNERS_ALL = 28
    TELNET_SAME_USER_PASS_DICTIONARY_ALL = 29
    SSH_SAME_USER_PASS_DICTIONARY_ALL = 30
    FTP_SAME_USER_PASS_DICTIONARY_ALL = 31
    CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL = 32
    IRC_SAME_USER_PASS_DICTIONARY_ALL = 33
    MONGO_SAME_USER_PASS_DICTIONARY_ALL = 34
    MYSQL_SAME_USER_PASS_DICTIONARY_ALL = 35
    SMTP_SAME_USER_PASS_DICTIONARY_ALL = 36
    POSTGRES_SAME_USER_PASS_DICTIONARY_ALL = 37
    NETWORK_SERVICE_LOGIN = 38
    FIND_FLAG = 39
    NIKTO_WEB_HOST_SCAN = 40
    MASSCAN_HOST_SCAN = 41
    MASSCAN_ALL_SCAN = 42
    FIREWALK_HOST = 43
    FIREWALK_ALL = 44
    HTTP_ENUM_HOST = 45
    HTTP_ENUM_ALL = 46
    HTTP_GREP_HOST = 47
    HTTP_GREP_ALL = 48
    FINGER_HOST = 49
    FINGER_ALL = 50
    INSTALL_TOOLS = 51
    SSH_BACKDOOR = 52
    SAMBACRY_EXPLOIT = 53
    SHELLSHOCK_EXPLOIT = 54
    DVWA_SQL_INJECTION = 55
    CVE_2015_3306_EXPLOIT = 56
    CVE_2015_1427_EXPLOIT = 57
    CVE_2016_10033_EXPLOIT = 58
    CVE_2010_0426_PRIV_ESC = 59
    CVE_2015_5602_PRIV_ESC = 60
    STOP = 61
    CONTINUE = 62
