"""
Type of vulnerabilities in an emulation
"""

from enum import IntEnum


class VulnType(IntEnum):
    """
    Enum representing the different vulnerability types for generated containers
    """
    WEAK_PW = 0
    RCE = 1  # remote code execution
    SQL_INJECTION = 2
    PRIVILEGE_ESCALATION = 3
